X = "✗"
DOT = "."
CHECK = "✓"
ARROW = "→"
QUOTE = '"'
NEW_LINE = "\n"
ANIMATION_SLEEP = 1.5
MATCH = "reference"
CACHE = r"\$\((?P<reference>.*?)\)"
NO_EXCEPTION = (None, None, None)

USER_ENV_PATH = ".env*"
USER_PIPFILE_PATH = "Pipfile"
USER_PIPFILE_INSTALL_COMMAND = ["-m", "pip", "install"]
USER_CLIENT_MODULE_NAME = "local"
USER_CLIENT_PATH_PREFIXES = "awyes.*"
USER_CLIENT_PATH_SUFFIXES = (".py",)
USER_CONFIG_PATH_PREFIXES = "awyes.*"
USER_CONFIG_PATH_SUFFIXES = (".yaml", ".yml")
